import os
import re
from re import sub, subn
import bs4
from .subsearcher import BaseSubSearcher


class SubHDSubSearcher(BaseSubSearcher):
    """ SubHD 字幕搜索器(https://subhd.tv)
    """
    SUPPORT_LANGUAGES = ['zh_chs', 'zh_cht', 'en', 'zh_en']
    SUPPORT_EXTS = ['ass', 'srt']
    LANGUAGES_MAP = {
        '简体': 'zh_chs',
        '繁體': 'zh_cht',
        '英文': 'en',
        '双语': 'zh_en'
    }
    COMMON_LANGUAGES = ['英文', '简体', '繁体']

    API_URL = 'https://subhd.tv/search/'
    API_SUBTITLE_DOWNLOAD = '/ajax/down_ajax'
    API_SUBTITLE_PREVIEW = '/ajax/file_ajax'

    _cache = {}
    shortname = 'subhd'

    def __init__(self, subfinder, **kwargs):
        super(SubHDSubSearcher, self).__init__(subfinder, **kwargs)
        self.API_SUBTITLE_DOWNLOAD = self.api_urls.get(
            'subhd_api_subtitle_download', self.__class__.API_SUBTITLE_DOWNLOAD)
        self.API_SUBTITLE_PREVIEW = self.api_urls.get(
            'subhd_api_subtitle_preview', self.__class__.API_SUBTITLE_DOWNLOAD
        )

    def _parse_search_results_html(self, doc):
        """ parse search result html
        """
        soup = bs4.BeautifulSoup(doc, 'lxml')
        subinfo_list = []
        div_list = soup.select('div.mb-4')
        if not div_list:
            return subinfo_list
        for div in div_list:
            subinfo = {
                'title': '',
                'link': '',
                'author': '',
                'exts': [],
                'languages': [],
                'rate': 0,
                'download_count': 0,
            }
            div_title = div.find('div', class_='f12 pt-1')
            if not div_title:
                break
            a = div_title.a
            # 字幕标题
            subinfo['title'] = a.get('title').strip()
            # 链接
            subinfo['link'] = a.get('href').strip()

            div_format = div_title.find_next_siblings('div', limit=1)
            if not div_format:
                break
            div_format = div_format[0]
            # 语言
            format_str = ' '.join(div_format.strings)
            for l1, l2 in self.LANGUAGES_MAP.items():
                if l1 in format_str:
                    subinfo['languages'].append(l2)
            # 格式
            for ext in self.SUPPORT_EXTS:
                if ext in format_str or ext.upper() in format_str:
                    subinfo['exts'].append(ext)
            # 下载次数
            div_download = div_format.find_next_siblings('div', class_='pt-3')
            if not div_download:
                break
            div_download = div_download[0]
            fa_download = div_download.find('i', class_='fa-download')
            dl_str = fa_download.next_sibling
            dl_str = dl_str.replace('次', '')
            subinfo['download_count'] = int(dl_str)
            subinfo_list.append(subinfo)
        return subinfo_list

    def _get_subinfo_list(self, videoname):
        """ return subinfo_list of videoname
        """
        # searching subtitles
        url = self._join_url(self.API_URL, videoname)
        res = self.session.get(url)
        doc = res.content
        referer = res.url
        subinfo_list = self._parse_search_results_html(doc)
        for subinfo in subinfo_list:
            subinfo['link'] = self._join_url(res.url, subinfo['link'])
        return subinfo_list, referer

    def _visit_detailpage(self, detailpage_link, referer):
        cookie = {
            'ci_session': '1bch2vfbae275t8cu8u9uqs0oks4gpk5',
            'Hm_lvt_36f45ef10337991c93242d418c95baa3': '1594656652',
            'Hm_lpvt_36f45ef10337991c93242d418c95baa3': '1594716149',
        }
        download_link = ''
        headers = {
            'Referer': referer
        }
        res = self.session.get(detailpage_link, headers=headers, cookies=cookie)
        if not res.ok:
            return download_link, ''
        doc = res.text
        referer = res.url
        soup = bs4.BeautifulSoup(doc, 'lxml')
        button_download = soup.find('button', id=True, sid=True)
        if not button_download:
            return download_link, ''
        api_subtitle_url = self._join_url(referer, self.API_SUBTITLE_DOWNLOAD)
        params = {
            'sub_id': button_download.get('sid'),
            'dtoken1': button_download.get('dtoken1'),
        }
        res = self.session.post(api_subtitle_url, json=params, cookies=cookie)
        if not res.ok:
            return download_link, ''
        data = res.json()
        if data['success']:
            download_link = data['url']
        else:
            self.subfinder.logger.info('遇到验证码, 请尝试手动下载: {}'.format(detailpage_link))
        return download_link, referer
    
    def _try_preview_subs(self, detailpage_link, referer, videofile):
        subs = []
        root = os.path.dirname(videofile)
        api_url = self._join_url(detailpage_link, self.API_SUBTITLE_PREVIEW)
        headers = {
            'Referer': referer
        }

        res = self.session.get(detailpage_link, headers=headers)
        if not res.ok:
            return subs
        doc = res.text
        referer = res.url
        soup = bs4.BeautifulSoup(doc, 'lxml')
        a_list = soup.select('a[data-target="#fileModal"][data-sid]')
        if not a_list:
            return subs
        files = []
        for a in  a_list:
            s = a.string.strip()
            if s == '预览':
                files.append((a.get('data-sid'), a.get('data-fname')))
        
        for sid, fname in files:
            params = {'dasid': sid, 'dafname': fname}
            resp = self.session.post(api_url, data=params)
            if not resp.ok:
                continue
            data = resp.json()
            if not data['success']:
                continue
            filedata = data['filedata']
            origin_file = os.path.basename(fname)
            subname = self._gen_subname(videofile, origin_file)
            subname = os.path.join(root, subname)
            with open(subname, 'w') as fp:
                fp.write(filedata)
            subs.append(subname)

        return subs

    def _search_subs(self, videofile, languages, exts, keyword=None):
        videoname = self._get_videoname(videofile)  # basename, not include ext
        videoinfo = self._parse_videoname(videoname)
        if keyword is None:
            keyword = self._gen_keyword(videoinfo)

        self._debug('keyword: {}'.format(keyword))
        self._debug('videoinfo: {}'.format(videoinfo))

        # try find subinfo_list from self._cache
        if keyword not in self._cache:
            subinfo_list, referer = self._get_subinfo_list(keyword)
            self._cache[keyword] = (subinfo_list, referer)
        else:
            subinfo_list, referer = self._cache.get(keyword)
        self._debug('subinfo_list: {}'.format(subinfo_list))

        subinfo = self._filter_subinfo_list(
            subinfo_list, videoinfo, languages, exts)
        self._debug('subinfo: {}'.format(subinfo))
        if not subinfo:
            return []

        subtitle_download_link, referer = self._visit_detailpage(
            subinfo['link'], referer)
        self._debug('subtitle_download_link: {}'.format(subtitle_download_link))
        
        subs = None
        if not subtitle_download_link:
            subs = self._try_preview_subs(subinfo['link'], referer, videofile)
        else:
            filepath, referer = self._download_subs(
                subtitle_download_link, videofile, referer, subinfo['title'])
            self._debug('filepath: {}'.format(filepath))
            subs = self._extract(filepath, videofile, exts)

        self._debug('subs: {}'.format(subs))

        return [{
            'link': referer,
            'language': subinfo['languages'],
            'ext': subinfo['exts'],
            'subname': subs,
            'downloaded': True
        }] if subs else []



# -*- coding: utf8 -*-
from __future__ import unicode_literals
import os
import sys
import glob
import fnmatch
import logging
import mimetypes
import traceback
import requests
from .subsearcher import get_subsearcher, exceptions


class Pool(object):
    """ 模拟线程池，实际上还是同步执行代码
    """

    def __init__(self, size):
        self.size = size

    def spawn(self, fn, *args, **kwargs):
        fn(*args, **kwargs)

    def join(self):
        return


class SubFinder(object):
    """ 字幕查找器
    """

    DEFAULT_VIDEO_EXTS = {'.mkv', '.mp4', '.ts', '.avi', '.wmv'}

    def __init__(self, path='./', languages=None, exts=None, subsearcher_class=None, **kwargs):
        self.set_path(path)
        self.languages = languages
        self.exts = exts
        self.subsearcher = []

        # silence: dont print anything
        self.silence = kwargs.get('silence', False)
        # logger's output
        self.logger_output = kwargs.get('logger_output', sys.stdout)
        # debug
        self.debug = kwargs.get('debug', False)
        # video_exts
        self.video_exts = set(self.__class__.DEFAULT_VIDEO_EXTS)
        if 'video_exts' in kwargs:
            video_exts = set(kwargs.get('video_exts'))
            self.video_exts.update(video_exts)
        # keyword
        self.keyword = kwargs.get('keyword')
        # ignore
        self.ignore = kwargs.get('ignore', False)
        # exclude
        self.exclude = kwargs.get('exclude', [])

        self._init_session()
        self._init_pool()
        self._init_logger()

        # _history: recoding downloading history
        self._history = {}

        if subsearcher_class is None:
            subsearcher_class = get_subsearcher('default')
        if not isinstance(subsearcher_class, list):
            subsearcher_class = [subsearcher_class]

        # api urls
        api_urls = kwargs.get('api_urls', {})
        
        for sc in subsearcher_class:
            self.subsearcher.append(sc(self, api_urls=api_urls))

    def _is_videofile(self, f):
        """ determine whether `f` is a valid video file, mostly base on file extension 
        """
        if os.path.isfile(f):
            types = mimetypes.guess_type(f)
            mtype = types[0]
            if (mtype and mtype.split('/')[0] == 'video') or (os.path.splitext(f)[1] in self.video_exts):
                return True
        return False

    def _has_subtitles(self, f):
        """ 判断f是否已经有了本地字幕
        """
        dirname = os.path.dirname(f)
        basename = os.path.basename(f)
        basename_no_ext, _ = os.path.splitext(basename)
        exts = self.exts or ['ass', 'srt']
        for filename in os.listdir(dirname):
            _, ext = os.path.splitext(filename)
            ext = ext[1:]
            if filename.startswith(basename_no_ext) and ext in exts:
                return True
        return False

    def _fnmatch(self, f):
        for pattern in self.exclude:
            if fnmatch.fnmatchcase(f, pattern):
                return True
        return False

    def _filter_path(self, path):
        """ 筛选出 path 目录下所有的视频文件
        """
        if self._is_videofile(path):
            if not self.ignore and self._has_subtitles(path):
                return
            yield path
            return

        if not os.path.isdir(path):
            return

        for root, dirs, files in os.walk(path):
            for filename in files:
                filepath = os.path.join(root, filename)
                if not self._is_videofile(filepath):
                    continue
                if self._fnmatch(filename):
                    continue
                if not self.ignore and self._has_subtitles(filepath):
                    continue
                yield filepath

            # remove dir in self.exclude
            removed_index = []
            for i, dirname in enumerate(dirs):
                if self._fnmatch(dirname + '/'):
                    removed_index.append(i)
            for i in removed_index:
                dirs.pop(i)

    def _init_session(self):
        """ 初始化 requests.Session
        """
        self.session = requests.Session()
        self.session.mount('http://', adapter=requests.adapters.HTTPAdapter(
            pool_connections=10,
            pool_maxsize=100))

    def _init_pool(self):
        self.pool = Pool(10)

    def _init_logger(self):
        log_level = logging.INFO
        if self.silence:
            log_level = logging.CRITICAL + 1
        if self.debug:
            log_level = logging.DEBUG
        self.logger = logging.getLogger('SubFinder')
        self.logger.handlers = []
        self.logger.setLevel(log_level)
        sh = logging.StreamHandler(stream=self.logger_output)
        sh.setLevel(log_level)
        formatter = logging.Formatter(
            '[%(asctime)s]-[%(levelname)s]: %(message)s', datefmt='%m/%d %H:%M:%S')
        sh.setFormatter(formatter)
        self.logger.addHandler(sh)

    def _download(self, videofile):
        """ 调用 SubSearcher 搜索并下载字幕
        """
        basename = os.path.basename(videofile)

        subinfos = []
        for subsearcher in self.subsearcher:
            self.logger.info(
                '{0}：开始使用 {1} 搜索字幕'.format(basename, subsearcher))
            try:
                subinfos = subsearcher.search_subs(
                    videofile, self.languages, self.exts, self.keyword)
            except Exception as e:
                err = str(e)
                if self.debug:
                    err = traceback.format_exc()
                self.logger.error(
                    '{}：搜索字幕发生错误： {}'.format(basename, err))
                continue
            if subinfos:
                break
        self.logger.info('{1}：找到 {0} 个字幕, 准备下载'.format(
            len(subinfos), basename))
        try:
            for subinfo in subinfos:
                downloaded = subinfo.get('downloaded', False)
                if downloaded:
                    if isinstance(subinfo['subname'], (list, tuple)):
                        self._history[videofile].extend(subinfo['subname'])
                    else:
                        self._history[videofile].append(subinfo['subname'])
                else:
                    link = subinfo.get('link')
                    subname = subinfo.get('subname')
                    subpath = os.path.join(os.path.dirname(videofile), subname)
                    res = self.session.get(link, stream=True)
                    with open(subpath, 'wb') as fp:
                        for chunk in res.iter_content(8192):
                            fp.write(chunk)
                    self._history[videofile].append(subpath)
        except Exception as e:
            self.logger.error(str(e))

    def set_path(self, path):
        path = os.path.abspath(path)
        self.path = path

    def start(self):
        """ SubFinder 入口，开始函数
        """
        self.logger.info('开始')
        videofiles = list(self._filter_path(self.path))
        l = len(videofiles)
        if l > 1 and self.keyword:
            self.logger.warn('`keyword` should used only when there is one video file, but there is {} video files'.format(l))
            return
        for f in videofiles:
            self._history[f] = []
            self.pool.spawn(self._download, f)
        self.pool.join()
        self.logger.info('='*20 + '下载完成' + '='*20)
        for v, subs in self._history.items():
            basename = os.path.basename(v)
            self.logger.info(
                '{}: 下载 {} 个字幕'.format(basename, len(subs)))

    def done(self):
        pass



# -*- coding: utf8 -*-
from __future__ import unicode_literals
from abc import abstractmethod, ABCMeta
import os
import re
import cgi
try:
    import urlparse
except ImportError as e:
    from urllib import parse as urlparse
import requests
from subfinder.tools.compressed_file import CompressedFile
from . import exceptions


registered_subsearchers = {}


def register_subsearcher(name, subsearcher_cls):
    """ register a subsearcher, the `name` is a key used for searching subsearchers.
    if the subsearcher named `name` already exists, then it's will overrite the old subsearcher.
    """
    if not issubclass(subsearcher_cls, BaseSubSearcher):
        raise ValueError(
            '{} is not a subclass of BaseSubSearcher'.format(subsearcher_cls))
    registered_subsearchers[name] = subsearcher_cls


def register(subsearcher_cls=None, name=None):
    def decorator(subsearcher_cls):
        if name is None:
            _name = subsearcher_cls.__name__
        else:
            _name = name
        register_subsearcher(_name, subsearcher_cls)
        return subsearcher_cls
    return decorator(subsearcher_cls) if subsearcher_cls is not None else decorator


def get_subsearcher(name, default=None):
    return registered_subsearchers.get(name, default)


def get_all_subsearchers():
    return registered_subsearchers


class BaseSubSearcher(object):
    """ The abstract class for search subtitles.

    You must implement following methods:
    - search_subs
    """
    __metaclass__ = ABCMeta

    SUPPORT_LANGUAGES = []
    SUPPORT_EXTS = []

    def __init__(self, subfinder,  **kwargs):
        """
        subfinder: SubFinder
        debug: 是否输出调试信息
        """
        self.session = requests.session()
        self.session.headers['User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36'
        self.subfinder = subfinder
        self.api_urls = kwargs.get('api_urls', {})
        self.API_URL = self.api_urls.get(
            self.shortname, self.__class__.API_URL)

    def _debug(self, msg):
        self.subfinder.logger.debug(msg)

    @classmethod
    def _check_languages(cls, languages):
        for lang in languages:
            if lang not in cls.SUPPORT_LANGUAGES:
                raise exceptions.LanguageError(
                    '{} doesn\'t support "{}" language'.format(cls.__name__, lang))

    @classmethod
    def _check_exts(cls, exts):
        for ext in exts:
            if ext not in cls.SUPPORT_EXTS:
                raise exceptions.ExtError(
                    '{} doesn\'t support "{}" ext'.format(cls.__name__, ext))

    @classmethod
    def _join_url(cls, url, path):
        """ join absolute `url` and `path`(href)
        """
        return urlparse.urljoin(url, path)

    @classmethod
    def _get_videoname(cls, videofile):
        """parse the `videofile` and return it's basename
        """
        name = os.path.basename(videofile)
        name = os.path.splitext(name)[0]
        return name

    RE_SEASON = re.compile(
        r'[Ss](?P<season>\d+)\.?')
    RE_SEASON_EPISODE = re.compile(
        r'[Ss](?P<season>\d+)\.?[Ee](?P<episode>\d+)')
    RE_RESOLUTION = re.compile(r'(?P<resolution>720[Pp]|1080[Pp]|2160[Pp]|HR)')
    RE_SOURCE = re.compile(
        r'\.(?P<source>BD|Blu[Rr]ay|BDrip|WEB-DL|HDrip|HDTVrip|HDTV|HD|DVDrip)\.')
    RE_AUDIO_ENC = re.compile(
        r'(?P<audio_encoding>mp3|DD5\.1|DDP5\.1|AC3\.5\.1)')
    RE_VIDEO_ENC = re.compile(r'(?P<video_encoding>x264|H\.264|AVC1|H\.265)')

    @classmethod
    def _parse_videoname(cls, videoname):
        """ parse videoname and return video info dict
        video info contains:
        - title, the name of video
        - sub_title, the sub_title of video
        - resolution,
        - source,
        -
        - season, defaults to 0
        - episode, defaults to 0
        """
        info = {
            'title': '',
            'season': 0,
            'episode': 0,
            'resolution': '',
            'source': '',
            'audio_encoding': '',
            'video_encoding': '',
        }
        mapping = {
            'resolution': cls.RE_RESOLUTION,
            'source': cls.RE_SOURCE,
            'audio_encoding': cls.RE_AUDIO_ENC,
            'video_encoding': cls.RE_VIDEO_ENC
        }
        index = len(videoname)
        m = cls.RE_SEASON_EPISODE.search(videoname)
        if m:
            info['season'] = int(m.group('season'))
            info['episode'] = int(m.group('episode'))
            index, _ = m.span()
            info['title'] = videoname[0:index].strip('.')
        else:
            m = cls.RE_SEASON.search(videoname)
            if m:
                info['season'] = int(m.group('season'))
                index, _ = m.span()
                info['title'] = videoname[0:index].strip('.')
        
        for k, r in mapping.items():
            m = r.search(videoname)
            if m:
                info[k] = m.group(k)
                i, e = m.span()
                if info['title'] == '' or i < index:
                    index = i
                    info['title'] = videoname[0:index].strip('.')

        if info['title'] == '':
            i = videoname.find('.')
            info['title'] = videoname[:i] if i > 0 else videoname

        return info
    
    @classmethod
    def _gen_keyword(cls, videoinfo):
        """ 获取关键词
        """
        keyword = videoinfo.get('title')
        if videoinfo['season'] != 0:
            keyword += '.S{:02d}'.format(videoinfo['season'])
        if videoinfo['episode'] != 0:
            keyword += '.E{:02d}'.format(videoinfo['episode'])
        # replace space with '+'
        keyword = re.sub(r'\s+', ' ', keyword)
        return keyword

    @classmethod
    def _gen_subname(cls, videofile, origin_file, language=None, ext=None):
        if not language:
            language_ = []
            try:
                for l in cls.COMMON_LANGUAGES:
                    if origin_file.find(l) >= 0:
                        language_.append(l)
            except Exception:
                pass
            language = '&'.join(language_)
        if language and not language.startswith('.'):
            language = '.' + language

        basename = os.path.basename(videofile)
        basename, _ = os.path.splitext(basename)
        if not ext:
            _, ext = os.path.splitext(origin_file)
        if not ext.startswith('.'):
            ext = '.' + ext

        return '{basename}{language}{ext}'.format(
            basename=basename,
            language=language,
            ext=ext)

    @classmethod
    def _extract(cls, compressed_file, videofile, exts):
        """ 解压字幕文件，如果无法解压，则直接返回 compressed_file。
        exts 参数用于过滤掉非字幕文件，只有文件的扩展名在 exts 中，才解压该文件。
        """
        if not CompressedFile.is_compressed_file(compressed_file):
            return [compressed_file]

        root = os.path.dirname(compressed_file)
        subs = []
        cf = CompressedFile(compressed_file)
        for name in cf.namelist():
            if cf.isdir(name):
                continue
            # make `name` to unicode string
            origin_file = CompressedFile.decode_file_name(name)
            _, ext = os.path.splitext(origin_file)
            ext = ext[1:]
            if ext not in exts:
                continue
            subname = cls._gen_subname(videofile, origin_file)
            subpath = os.path.join(root, subname)
            cf.extract(name, subpath)
            subs.append(subpath)
        cf.close()
        return subs

    @classmethod
    def _filter_subinfo_list(cls, subinfo_list, videoinfo, languages, exts):
        """ filter subinfo list base on:
        - season
        - episode
        - languages
        - exts
        -
        return a best matched subinfo
        """
        filter_field_list = [
            'season',
            'episode',
            'resolution',
            'source',
            'video_encoding',
            'audio_encoding'
        ]
        filtered_subinfo_list = dict((f, []) for f in filter_field_list)

        for subinfo in subinfo_list:
            title = subinfo.get('title')
            videoinfo_ = cls._parse_videoname(title)
            last_field = None
            for field in filter_field_list:
                i = videoinfo.get(field)
                if isinstance(i, str):
                    i = i.lower()
                j = videoinfo_.get(field)
                if isinstance(j, str):
                    j = j.lower()
                if i == j:
                    last_field = field
                else:
                    break
            if last_field is not None:
                filtered_subinfo_list[last_field].append(subinfo)
        for field in filter_field_list[::-1]:
            if len(filtered_subinfo_list[field]) > 0:
                # sort by download_count and rate
                sorted_subinfo_list = sorted(filtered_subinfo_list[field],
                                             key=lambda item: (
                    item['rate'], item['download_count']),
                    reverse=True)
                return sorted_subinfo_list[0]
        return None

    def _download_subs(self, download_link, videofile, referer='', sub_title=''):
        """ 下载字幕
        videofile: 视频文件路径
        sub_title: 字幕标题（文件名）
        download_link: 下载链接
        referer: referer
        """
        root = os.path.dirname(videofile)
        name, _ = os.path.splitext(os.path.basename(videofile))
        ext = ''

        headers = {
            'Referer': referer
        }
        res = self.session.get(download_link, headers=headers, stream=True)
        referer = res.url

        # 尝试从 Content-Disposition 中获取文件后缀名
        content_disposition = res.headers.get('Content-Disposition', '')
        if content_disposition:
            _, params = cgi.parse_header(content_disposition)
            filename = params.get('filename')
            if filename:
                _, ext = os.path.splitext(filename)
                ext = ext[1:]

        if ext == '':
            # 尝试从url 中获取文件后缀名
            p = urlparse.urlparse(res.url)
            path = p.path
            if path:
                _, ext = os.path.splitext(path)
                ext = ext[1:]

        if ext == '':
            # 尝试从字幕标题中获取文件后缀名
            _, ext = os.path.splitext(sub_title)
            ext = ext[1:]

        filename = '{}.{}'.format(name, ext)
        filepath = os.path.join(root, filename)
        with open(filepath, 'wb') as fp:
            for chunk in res.iter_content(8192):
                fp.write(chunk)

        return filepath, referer

    def _search_subs(self, videofile, languages, exts, keyword=None):
        """ search subtitles of videofile.

        `videofile` is the absolute(or relative) path of the video file.

        `languages` is the language of subtitle, e.g chn, eng, the support for language is difference, depende on
        implemention of subclass. `languages` accepts one language or a list of language

        `exts` is the format of subtitle, e.g ass, srt, sub, idx, the support for ext is difference,
        depende on implemention of subclass. `ext` accepts one ext or a list of ext

        `keyword` is used to searching on the subtitle website.

        return a list of subtitle info
        [
            {
                'link': '',     # download link
                'language': '', # language
                'ext': '',      # ext
                'subname': '',  # the filename of subtitles
                'downloaded': False, # it's tell `SubFinder` whether need to download.
            },
            {
                'link': '',
                'language': '',
                'ext': '',
                'subname': '',
            },
            ...
        ]
        - `link`, it's optional, but if `downloaded` is False, then `link` is required.
        - `language`, it's optional
        - `subname`, it's optional, but if `downloaded` is False, then `subname` is required.
        - `downloaded`, `downloaded` is required.
            if `downloaded` is True, then `SubFinder` will not download again,
            otherwise `SubFinder` will download link.

        sub-class should implement this private method.
        """
        return []

    def search_subs(self, videofile, languages=None, exts=None, keyword=None):
        if languages is None:
            languages = self.SUPPORT_LANGUAGES
        elif isinstance(languages, str):
            languages = [languages]
        self._check_languages(languages)

        if exts is None:
            exts = self.SUPPORT_EXTS
        elif isinstance(exts, str):
            exts = [exts]
        self._check_exts(exts)

        return self._search_subs(videofile, languages, exts, keyword)

    def __str__(self):
        if hasattr(self.__class__, 'shortname'):
            name = self.__class__.shortname
        else:
            name = self.__class__.__name__
        return '<{}>'.format(name)

    def __unicode__(self):
        return self.__str__()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['subfinder', 'subfinder.subsearcher', 'subfinder.tools']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2',
 'gevent>=1.4.0,<2.0.0',
 'lxml',
 'rarfile',
 'requests',
 'six']

entry_points = \
{'console_scripts': ['subfinder = subfinder.run_gevent:run',
                     'subutils = subfinder.utils:main']}

setup_kwargs = {
    'name': 'subfinder',
    'version': '1.1.3',
    'description': 'subfinder -- a general finder for subtitles',
    'long_description': '# subfinder 字幕查找器\n\nsubfinder 是一个通用字幕查找器，可以查找字幕并下载。\n\n---\n\n[![PyPi](https://img.shields.io/pypi/v/subfinder.svg)](https://pypi.org/project/subfinder/)\n[![PyPi - Python Version](https://img.shields.io/pypi/pyversions/subfinder.svg)](https://pypi.org/project/subfinder/)\n[![PyPI - License](https://img.shields.io/pypi/l/subfinder.svg)](https://pypi.org/project/subfinder/)\n![GitHub last commit](https://img.shields.io/github/last-commit/ausaki/subfinder.svg)\n\n---\n\nTable of Contents\n=================\n\n  - [特性](#特性)\n  - [安装](#安装)\n  - [更新](#更新)\n  - [使用方法](#使用方法)\n    - [命令行](#命令行)\n    - [配置文件](#配置文件)\n    - [Windows 右键菜单](#windows-右键菜单)\n    - [macOS 右键菜单](#macos-右键菜单)\n  - [图形界面](#图形界面)\n  - [注意事项](#注意事项)\n    - [shooter 字幕搜索器](#shooter-字幕搜索器)\n    - [zimuku 字幕搜索器](#zimuku-字幕搜索器)\n  - [贡献](#贡献)\n  - [参考](#参考)\n  - [License](#license)\n  - [更新历史](#更新历史)\n    - [v1.1.3](#v113)\n    - [v1.1.2](#v112)\n    - [v1.1.1](#v111)\n    - [v1.1.0](#v110)\n    - [v1.0.9](#v109)\n    - [v1.0.8](#v108)\n    - [v1.0.7](#v107)\n    - [v1.0.6](#v106)\n    - [v1.0.5](#v105)\n    - [v1.0.4](#v104)\n    - [v1.0.3](#v103)\n    - [v1.0.2](#v102)\n    - [v1.0.1](#v101)\n\n---\n\n## 特性\n\n- 支持提供的 API，可以精确匹配字幕。\n\n- 支持 [射手字幕网](https://www.shooter.cn/), [字幕库](https://www.zimuku.cn/), [字幕组](http://www.zimuzu.io/), [SubHD](https://subhd.tv)。\n\n## 安装\n\n如果你是 Linux 和 macOS 用户，由于系统自带 Python，推荐使用下面的方法安装 subfinder：\n\n`pip install subfinder`\n\n安装完成之后，会在 Python 的 scripts 目录下添加一个叫做 subfinder 的可执行文件。\n\n> 在 unix-like 系统中，scripts 目录一般是 `/usr/local/bin`，在 Windows 系统中，scripts 目录一般是 `C:\\python\\scripts\\`。在 Windows 系统中需要将 `C:\\python\\scripts\\` 加入到 `PATH` 中（一般安装 Python 时已经添加了）。\n\n接下来你就可以在命令行中使用 subfinder 命令了。\n\n为了方便没有安装 Python 的用户，这里也有已经打包好的可执行文件。[戳这里下载](https://github.com/ausaki/subfinder/releases)。\n\n## 更新\n\n如果 subfinder 是使用 pip 安装的，那么使用 pip 更新，\n\n`pip install subfinder --upgrade`\n\n如果 subfinder 是下载的打包好的可执行文件，那么请重新下载最新的可执行文件并覆盖旧的文件。\n\n## 使用方法\n\n### 命令行\n\n- 使用默认字幕查找器（shooter）查找单个视频的字幕：\n\n  `subfinder /path/to/videofile`\n\n- 使用默认字幕查找器（shooter）查找目录下（递归所有子目录）所有视频的字幕：\n\n  `subfinder /path/to/directory_contains_video`\n\n- 使用指定的字幕查找器查找字幕，例如 zimuku：\n\n  `subfinder /path/to/videofile -m zimuku`\n\n- 同时使用多个字幕查找器查找字幕\n\n  `subfinder /path/to/videofile -m shooter zimuku`\n\n  当指定多个字幕查找器时，subfinder 会依次尝试每个字幕查找器去查找字幕，只要有一个字幕查找器返回字幕信息，则不再使用后面的字幕查找器查找字幕。\n\n  ** 注意：** 如果指定了多个字幕查找器，请不要指定 `languages` 参数，否则可能会出现 `LanguageError` 错误（因为每个 `SubSearcher` 支持的语言可能不相同）。\n\n常用参数说明（详细的参数信息请查看 `subfinder -h`）：\n\n| 参数              | 含义                                                                                               | 必需                                               |\n| ----------------- | -------------------------------------------------------------------------------------------------- | -------------------------------------------------- |\n| `-l, --languages` | 指定字幕语言，可同时指定多个。每个字幕查找器支持的语言不相同。具体支持的语言请看下文。             | 否，subfinder 默认会下载字幕查找器找到的所有字幕。 |\n| `-e, --exts`      | 指定字幕文件格式，可同时指定多个。每个字幕查找器支持的文件格式不相同。具体支持的文件格式请看下文。 | 否，subfinder 默认会下载字幕查找器找到的所有字幕。 |\n| `-m,--method`     | 指定字幕查找器，可同时指定多个。                                                                   | 否，subfinder 默认使用 shooter 查找字幕。          |\n| `-k, --keyword`   | 手动搜索关键字. 当 SubFinder 使用本身的关键字无法搜索到字幕时, 可以通过这个参数手动指定关键字.     | 否                                                 |\n| `--video_exts`    | 视频文件的后缀名（包括.，例如. mp4）                                                               | 否                                                 |\n| `--ignore`        | 忽略本地已有的字幕强行查找字幕. 注意: 这可能会覆盖本地已有的字幕. 默认 False。                     | 否                                                 |\n| `--exclude`       | 排除文件或目录，支持类似于 shell 的文件匹配模式。详情见下文                                        | 否                                                 |\n| `--api_urls`      | 指定字幕搜索器的 API URL。详情见下文                                                               | 否                                                 |\n| `-c, --conf`      | 配置文件                                                                                           | 否，SubFinder 默认从~/.subfinder.json 读取。       |\n| `-s,--silence`    | 静默运行，不输出日志                                                                               | 否                                                 |\n| `--debug`         | 调试模式，输出调试日志                                                                             | 否                                                 |\n| `-h,--help`       | 显示帮助信息                                                                                       | 否                                                 |\n\n- `--exclude`, 支持的匹配模式类似于 shell，`*` 匹配任意长度的字符串，`?` 匹配一个字符，`[CHARS]` 匹配 CHARS 中的任一字符。例如：\n\n   - 排除包含 `abc` 的目录：`--exclude \'*abc*/\'`。注意添加单引号，防止 shell 对其进行扩展。\n\n   - 排除包含 `abc` 的文件：`--exclude \'*abc*\'`。注意和上个例子的区别，匹配目录时结尾有 `/` 目录分隔符，匹配文件则没有。\n\n\n- `--api_urls`\n\n  [字幕库](http://www.zimuku.la) 的链接不太稳定，有时候会更换域名，因此提供 `--api_urls` 选项自定义 API URL，以防域名或链接变动。\n\n  `--api_urls` 只接收 JSON 格式的字符串。\n\n  获取正确的 API URL 的方法：\n\n  - 字幕库的 API 一般形如 http://www.zimuku.la/search， 这个 URL 就是网页端 “搜索” 功能的 URL。\n\n  - 字幕组的 API 一般形如 http://www.zmz2019.com/search， 这个 URL 同样是网页端 “搜索” 功能的 URL。\n\n  - SubHD 的 API 一般形如 https://subhd.tv/search. \n\n  - 射手网的 API 比较稳定，一般不会变动。\n\n  **如果发现字幕网站的 API URL 发生改变, 欢迎提交 issue.**\n\n  配置示例：\n\n  ```\n  {\n    // 设置字幕库的搜索 API\n    "zimuku": "http://www.zimuku.la/search",\n    // 设置字幕组的搜索 API\n    "zimuzu": "http://www.zmz2019.com/search",\n    // 设置字幕组获取字幕下载链接的 API, 注意不包含域名\n    "zimuzu_api_subtitle_download": "/api/v1/static/subtitle/detail",\n    // 设置 SubHD 的搜索 API\n    "subhd": "https://subhd.tv/search",\n    // 设置 SubHD 获取字幕下载链接的 API, 注意不包含域名\n    "subhd_api_subtitle_download": "/ajax/down_ajax",\n    // 设置 SubHD 获取字幕预览的 API, 注意不包含域名\n    "subhd_api_subtitle_preview": "/ajax/file_ajax"\n  }\n  ```\n\n支持的语言和文件格式：\n\n| 字幕查找器 | 语言                                | 文件格式       |\n| ---------- | ----------------------------------- | -------------- |\n| shooter    | [\'zh\', \'en\']                        | [\'ass\', \'srt\'] |\n| zimuku     | [\'zh_chs\', \'zh_cht\', \'en\', \'zh_en\'] | [\'ass\', \'srt\'] |\n| zimuzu     | [\'zh_chs\', \'zh_cht\', \'en\', \'zh_en\'] | [\'ass\', \'srt\'] |\n| subhd      | [\'zh_chs\', \'zh_cht\', \'en\', \'zh_en\'] | [\'ass\', \'srt\'] |\n\n语言代码：\n\n| 代码   | 含义               |\n| ------ | ------------------ |\n| zh     | 中文，简体或者繁体 |\n| en     | 英文               |\n| zh_chs | 简体中文           |\n| zh_cht | 繁体中文           |\n| zh_en  | 双语               |\n\n### 配置文件\n\n配置文件是 JSON 格式的，支持命令行中的所有选项。命令行中指定的选项优先级高于配置文件的。\n\n配置文件中的 key 一一对应于命令行选项，例如 `-m，--method` 对应的 key 为 `method`。\n\n示例：\n\n```json\n{\n  "languages": ["zh", "en", "zh_chs"],\n  "exts": ["ass", "srt"],\n  "method": ["shooter", "zimuzu", "zimuku"],\n  "video_exts": [".mp4", ".mkv", ".iso"],\n  "exclude": ["excluded_path/", "*abc.mp4"],\n  "api_urls": {\n    // 设置字幕库的搜索 API\n    "zimuku": "http://www.zimuku.la/search",\n    // 设置字幕组的搜索 API\n    "zimuzu": "http://www.zmz2019.com/search",\n    // 设置字幕组获取字幕下载链接的 API, 注意不包含域名\n    "zimuzu_api_subtitle_download": "/api/v1/static/subtitle/detail",\n    // 设置 SubHD 的搜索 API\n    "subhd": "https://subhd.tv/search",\n    // 设置 SubHD 获取字幕下载链接的 API, 注意不包含域名\n    "subhd_api_subtitle_download": "/ajax/down_ajax",\n    // 设置 SubHD 获取字幕预览的 API, 注意不包含域名\n    "subhd_api_subtitle_preview": "/ajax/file_ajax"\n  }\n}\n```\n\n### Windows 右键菜单\n\n使用命令行下载字幕还是有一点不方便，特别是需要输入路径。\n\n在 widnows 中，可以通过注册表将 subfinder 添加到右键菜单，使用时右键选中视频文件或者文件夹，然后点击右键菜单中的 “查找字幕”。\n\n![win_menu](assets/subfinder_menu.win.gif)\n\n** 如何添加注册表：**\n\n- 下载 [注册表文件](https://raw.githubusercontent.com/ausaki/subfinder/master/assets/subfinder.reg)，\n\n- 双击注册表文件 subfinder.reg 即可添加注册表到系统中。\n\n### macOS 右键菜单\n\n在 macOS 中，通过 Automator 的 Service 实现类似于 Windows 中的右键菜单功能。\n\n![subfinder_workflow_service.macos](assets/subfinder_workflow_service.macos.png)\n\n使用方法：\n\n- [下载 workflow](https://raw.githubusercontent.com/ausaki/subfinder/master/assets/subfinder.workflow.tar.gz)。\n- 解压 subfinder.workflow.tar.gz。\n- 将解压出的 subfinder.workflow 复制到 / Users/YourName/Library/Services。\n- 选中视频文件或目录，右键弹出菜单，选择 “服务（Services）” -> “查找字幕”。\n\n** 注意：在 workflow 中，subfinder 的路径是 `/usr/local/bin/subfinder`。**\n\n如果想要了解如何配置 workerflow，可以参考:\n\n- [macOS Automator 帮助](https://support.apple.com/zh-cn/guide/automator/welcome/mac)\n- [stackexchange 的这篇回答](https://apple.stackexchange.com/questions/238948/osx-how-to-add-a-right-click-option-in-folder-to-open-the-folder-with-an-applic?utm_medium=organic&utm_source=google_rich_qa&utm_campaign=google_rich_qa)\n\n## 图形界面\n\n打包好的可执行文件其实是一个 GUI app，不带任何参数运行的话，会打开一个图形界面。\n\n![subfinder_gui.win](assets/subfinder_gui.win.png)\n\n** 注意 **\n\n- GUI app 默认同时使用 shooter 和 zimuku 两个字幕搜索器。\n\n- GUI app 同样支持命令行参数。\n\n[下载页面](https://github.com/ausaki/subfinder/releases)\n\n## 注意事项\n\n### shooter 字幕搜索器\n\n- 由于射手字幕网爬虫的实时性，可能无法查找到最新发布视频的字幕。\n- 射手字幕网 API 返回的字幕可能出现 “语言不一致” 问题（指定查找英文字幕却返回中文字幕）。\n\n### zimuku 字幕搜索器\n\n- zimuku 网站明明有字幕可供下载，subfinder 却提示找不到字幕。可能的原因有：\n\n  - zimuku 字幕搜索器从视频文件名中提取的关键词不够准确，导致搜索结果为空。\n\n  - zimuku 网站修改了 HTML 代码，导致 HTML 解析失败。\n\n- 从 zimuku 网站上下载的字幕一般都是压缩包（zip, rar）。Python 自带的 zipfile 标准库可以解压 zip 压缩包，不过解压 rar 压缩包需要操作系统安装有 unrar 工具。由于 windows 系统安装和配置 unrar 比较麻烦，subfinder 已经內建 unrar.exe 了。对于 Linux 和 macOS 系统的用户，需要自己手动使用系统包管理工具安装 unrar。\n\n## 贡献\n\n在使用过程中遇到任何问题，请提交 issue。\n\n如果你希望分享你自己的字幕搜索器，欢迎提交 PR。\n\n## 参考\n\n- [射手字幕网 API 使用文档](https://docs.google.com/document/d/1ufdzy6jbornkXxsD-OGl3kgWa4P9WO5NZb6_QYZiGI0/preview)\n\n- [射手字幕网 API](https://www.shooter.cn/api/subapi.php)\n\n## License\n\n[MIT License](LICENSE)\n\n## 更新历史\n\n### v1.1.3\n\n- 将参数 `--repeat` 修改为 `--ignore`.\n\n- 添加新参数 `-k, --keyword`.\n\n- 支持 SubHD. SubHD 在下载字幕时经常弹出验证码, 无法通过正常的API获取到字幕的下载链接, 目前的做法是通过 SubHD 的字幕预览功能获取字幕.\n\n- 修复一些 bug.\n\n- 注意: 配置文件中的一些配置项修改了名字, 具体查看文档.\n  \n### v1.1.2\n\n- 修复 zimuzu 无法正常解析网页的问题 (因为 zimuzu 的网页结构发生了改变).\n\n### v1.1.1\n\n- 修复 subsearcher 的一个 bug.\n\n### v1.1.0\n\n- 使用 poetry 包管理器.\n\n- 修复若干 bug.\n\n### v1.0.9\n\n[issue 27](https://github.com/ausaki/subfinder/issues/27)\n\n- 支持用户配置文件\n\n- 支持用户自定义字幕组和字幕库的 API URL。\n\n- 支持用户自定义视频文件的后缀。\n\n- 支持忽略文件或文件夹。\n\n### v1.0.8\n\n- 修复 zimuku 搜索器的 bug。\n\n### v1.0.7\n\n- 升级 requests 至 2.20.0，修复 requests [安全问题](https://nvd.nist.gov/vuln/detail/CVE-2018-18074)，\n\n### v1.0.6\n\n- 优化了 subsearcher。\n\n- 添加 --debug 选项，方便显示调试信息。\n\n### v1.0.5\n\n- 修复命令行参数 `-l`，`-e` 的问题。\n\n### v1.0.4\n\n- 新增 zimuzu 字幕搜索器。\n\n- 优化代码。\n\n### v1.0.3\n\n- 优化 zimuku 字幕搜索器。\n\n- 內建 unrar.exe，windows 用户不用安装 unrara 了。\n\n### v1.0.2\n\n- GUI app 同时使用 shooter 和 zimuku 两个字幕搜索器搜索字幕。\n\n- GUI app 支持命令行运行。\n\n- 完善打包 GUI app 的流程。\n\n### v1.0.1\n\n- 完善 ZimukuSubsearcher。\n\n  - 解压字幕压缩包文件时，只解压字幕文件。字幕组上传的字幕压缩包文件中可能包含其它非字幕文件。\n\n  - 完善搜索功能。\n\n- 完善打包方式。\n\n- 修复一些 bug。\n',
    'author': 'ausaki',
    'author_email': 'ljm51689@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ausaki/subfinder/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)
